--[[
Name: LibGuildBankComm-1.0
Revision: $Revision: 26 $
Author(s): Myrroddin of Llane (Alliance US)
Description: Saves guild bank contents and sends/receives updates over the addon channel
Dependencies: LibStub, CallbackHandler-1.0, AceComm-3.0, AceSerializer-3.0, LibCompress
]]

local MAJOR_VERSION = "LibGuildBankComm-1.0"
local MINOR_VERSION = 90000 + tonumber(("$Revision: 26 $"):match("%d+"))

if not LibStub then error(MAJOR_VERSION.."requires LibStub") end
local lib, oldversion = LibStub:NewLibrary(MAJOR_VERSION, MINOR_VERSION)
if not lib then return end

local _G = getfenv(0)

LibStub("AceSerializer-3.0"):Embed(lib)
LibStub("AceComm-3.0"):Embed(lib)
local LibCompress = LibStub("LibCompress")

local function printError(action, err)
	print(format("%s - Error during %s: %s"), MAJOR_VERSION, action, err)
end

lib.eventFrame = lib.eventFrame or CreateFrame("Frame")
lib.eventFrame:SetScript("OnEvent", function(self, event, ...)
	self[event](self, ...)
end)
lib.eventFrame:UnregisterAllEvents()

local guildBank = {}
local noData = true
local dummy = {} --blank table to pass to pairs() when no real data is available
local playerName = UnitName("player")
local joinedGuild
local ranksDirty
local logged_in
local tabsChanged

lib.callbacks = lib.callbacks or LibStub("CallbackHandler-1.0"):New(lib)

function lib:OnCommReceived(prefix, msg, distribution, sender)
	if prefix ~= "LGBC" then return end
    if sender == playerName then return end
	
	local guildName = GetGuildInfo("player")
	local bankData
	
	local command, value = msg:match("^(%S+) (.*)$")
	
	if msg == "REQUEST" and not noData then
		local numTabs = guildBank.numTabs
		
		local commTabs = ("TABS %d"):format(numTabs)
		lib:SendCommMessage("LGBC", commTabs, "WHISPER", sender, "BULK")
		
		for page = 1, numTabs do
			local _, _, canView = GetGuildBankTabInfo(page)
			if canView then
				local serialized_data, serialization_error = lib:Serialize(guildBank[page])
				if serialization_error then
					return printError("serialization", serialization_error)
				end
					
				local compressed_data, compression_error = LibCompress:Compress(serialized_data)
				if compression_error then
					return printError("compression", compression_error)
				end
				
				-- Construct a comm string for the compressed data.
				local commString = ("UPDATE_PAGE_%d "):format(page) .. compressed_data
				lib:SendCommMessage("LGBC", commString, "WHISPER", sender, "BULK")
				--[===[@debug@
				print(("Sending page %d to %s"):format(page, sender))
				--@end-debug@]===]
			end
		end
		
		local commFunds = ("FUNDS %d"):format(guildBank.guildFunds)
		lib:SendCommMessage("LGBC", commFunds, "WHISPER", sender, "BULK")
	elseif command == "FUNDS" then
		local f = tonumber(value)
		guildBank.guildFunds = f
		lib.callbacks:Fire("GuildBankComm_FundsUpdate", sender, f, guildName)
		--[===[@debug@
		print("Funds callback fired")
		--@end-debug@]===]
	elseif command == "TABS" then
		local n = tonumber(value)
		guildBank.numTabs = n
		noData = false
		lib.callbacks:Fire("GuildBankComm_TabsUpdate", sender, n, guildName)
		--[===[@debug@
		print("Tabs callback fired")
		--@end-debug@]===]
	elseif command:find("^UPDATE_PAGE_%d+$") then
		local page = command:match("^UPDATE_PAGE_(%d+)$")
		local _, _, canView = GetGuildBankTabInfo(page)
		
		if canView then
			local decompressed_data, decompression_error = LibCompress:Decompress(value)
			if decompression_error then
				return printError("decompression", decompression_error)
			end
			
			local deserialized_data, deserialization_error = lib:Deserialize(decompressed_data)
			if deserialization_error then
				return printError("deserialization", deserialization_error)
			end
			
			guildBank[page] = deserialized_data
			noData = false
			lib.callbacks:Fire("GuildBankComm_PageUpdate", sender, page, guildName)
			--[===[@debug@
			print(("Page %d callback fired"):format(page))
			--@end-debug@]===]
		end
	end
end

lib:RegisterComm("LGBC")

function lib:RequestDataAboutPlayer()
	lib:SendCommMessage("LGBC", "REQUEST", "GUILD", nil, "BULK")
	
	if ranksDirty then
		-- forget any bank tab info the player should not have
		for page = 1, guildBank.numTabs do
			local _, _, canView = GetGuildBankTabInfo(page)
			if not canView then
				guildBank[page].links = nil
				guildBank[page].stacks = nil
			end
		end
	end
end

function lib.eventFrame:GUILD_ROSTER_UPDATE()
	-- if the player joins a guild
	-- or if the player is in a guild at PLAYER_LOGIN
	-- or is the player's rank changed
	-- we need to request new data
	
	if (joinedGuild) or (ranksDirty) or (logged_in) then
		lib:RequestDataAboutPlayer()
	end
	-- logged_in should be true only for PLAYER_LOGIN() and /reloadUI
	-- GUILD_ROSTER_UPDATE() and GUILD_RANKS_UPDATE() fire at same time when a unit's rank changes
	-- therefore reset playerRankChange in case player's rank changes rapidly and often
	ranksDirty = false
	logged_in = false
end

function lib.eventFrame:GUILD_RANKS_UPDATE()
	ranksDirty = true
end

function lib.eventFrame:PLAYER_GUILD_UPDATE()
	if not IsInGuild() then -- player left guild			
		wipe(guildBank)
		lib.callbacks:Fire("GuildBankComm_PlayerGuildless")
		joinedGuild = false			
		-- guild leaders and unguilded people both have rank 0 just to be confusing
		-- therefore we need to force the issue when leaving a guild
		ranksDirty = false
	else
		joinedGuild = true
		ranksDirty = true
	end
end

function lib.eventFrame:GUILDBANKFRAME_OPENED()
	local t = GetNumGuildBankTabs()
	if t ~= guildBank.numTabs then
		guildBank.numTabs = t
		tabsChanged = true
	end
	guildBank.guildFunds = GetGuildBankMoney()
	--[===[@debug@
	local tab = GetCurrentGuildBankTab()
	print("Guild Bank frame open. Looking at tab "..tostring(tab))
	tab = nil
	--@end-debug@]===]
	
	local name, icon
	for page = 1, guildBank.numTabs do
		name, icon = GetGuildBankTabInfo(page)
		
		if not guildBank[page] then
			guildBank[page] = {
				links = {},  --store the slot item links here
				stacks = {}, --store the slot stack counts here
				name = name,
				texture = icon
			}
		end
	end
	noData = false
	
	--register the _CLOSED event whenever the bank is opened
	self:RegisterEvent("GUILDBANKFRAME_CLOSED")
end

function lib.eventFrame:GUILDBANK_UPDATE_MONEY()
	guildBank.guildFunds = GetGuildBankMoney()
	--[===[@debug@
	local guildName = GetGuildInfo("player")
	print(guildName.." has "..guildBank.guildFunds.." in copper")
	--@end-debug@]===]
end

function lib.eventFrame:GUILDBANK_UPDATE_TABS()
	local t = GetNumGuildBankTabs()
	if t ~= guildBank.numTabs then
		guildBank.numTabs = t
		tabsChanged = true
		
		--if a new tab was purchased create the necessary 
		--storage structure
		if not guildBank[t] then
			local name, icon = GetGuildBankTabInfo(t)
			guildBank[t] = {
				links = {},
				stacks = {},
				name = name, 
				texture = icon
			}
		end
	end
end

local num, itemString
function lib.eventFrame:GUILDBANKBAGSLOTS_CHANGED()
	if (GuildBankFrame and GuildBankFrame:IsVisible()) or (BagnonFrameguildbank and BagnonFrameguildbank:IsVisible()) then
		local page = GetCurrentGuildBankTab()
		
		--98 slots on a page
		for slot = 1, 98 do
			--see if the slot has an item in it
			if GetGuildBankItemLink(page, slot) then
				--if there is link, parse the ID and place that in table w/ key = slot
				_, _, itemString = string.find(GetGuildBankItemLink(page, slot), "^|c%x+|H(.+)|h%[.+%]")
				guildBank[page].links[slot] = itemString
				
				--get item info too, need for stack sizes in image
				_, num = GetGuildBankItemInfo(page, slot)
				guildBank[page].stacks[slot] = num
			else
				-- that slot is empty
				guildBank[page].links[slot] = nil
				guildBank[page].stacks[slot] = nil
			end
			num, itemString = nil, nil
		end
		
		--save a timestamp for this scan
		guildBank[page].lastScan = time()
	end
end

function lib.eventFrame:GUILDBANKFRAME_CLOSED()
	--we only need the first of the two _CLOSED events
	self:UnregisterEvent("GUILDBANKFRAME_CLOSED")
	
	-- we want to filter what the player sees
	-- and only send that rather than the whole of lib.guildBank	
	local numTabs = guildBank.numTabs
	local guildName = GetGuildInfo("player")
	
	--only send a message about the number of tabs if it has changed
	if tabsChanged then
		local commTabs = ("TABS %d"):format(numTabs)
		lib:SendCommMessage("LGBC", commTabs, "GUILD", nil, "BULK")
		--[===[@debug@
		print("Sending number of tabs across comm")
		--@end-debug@]===]
		tabsChanged = false
	end
	
	--it is assumed the funds change often enough due to guild perks
	--that we should always send the latest value across
	local commFunds = ("FUNDS %d"):format(guildBank.guildFunds)
	lib:SendCommMessage("LGBC", commFunds, "GUILD", nil, "BULK")
	--[===[@debug@
	print("Sending guild funds across comm")
	--@end-debug@]===]
	
	for page = 1, numTabs do
		local _, _, canView = GetGuildBankTabInfo(page)
		if canView then
			local serialized_data, serialization_error = lib:Serialize(guildBank[page])
			if serialization_error then
				return printError("serialization", serialization_error)
			end
					
			local compressed_data, compression_error = LibCompress:Compress(serialized_data)
			if compression_error then
				return printError("compression", compression_error)
			end
				
			-- Construct a comm string for the compressed data.
			local commString = ("UPDATE_PAGE_%d "):format(page) .. compressed_data
			lib:SendCommMessage("LGBC", commString, "GUILD", nil, "BULK")
			--[===[@debug@
			print(("Sending page %d to Guild"):format(page))
			--@end-debug@]===]
		end
	end
end

function lib.eventFrame:PLAYER_LOGIN()
	logged_in = true
	if IsInGuild() then
		_G.GuildRoster()
	end
end

--Returns the current guild funds in copper
function lib:GetGuildFunds()
	return guildBank.guildFunds or 0
end

--Returns the current number of bank tabs owned by the guild
function lib:GetNumBankTabs()
	return guildBank.numTabs or 0
end

--Returns the specified tab's name and icon texture path
function lib:GetTabInfo(page)
	if type(page) ~= "number" then
		error("LibGuildBankComm:IterateLinks(page) expected number", 2)
	end
	if noData then return end
	if page < 1 or page > guildBank.numTabs then
		error(("LibGuildBankComm:GetTabInfo(page) : page must be between 1 and %d inclusive"):format(guildBank.numTabs), 2)
	end
	return guildBank[page].name, guildBank[page].texture
end

--Returns an iterator over both the links and stack counts for the slots of the given page
function lib:IteratePage(page)
	if type(page) ~= "number" then
		error("LibGuildBankComm:IterateLinks(page) expected number", 2)
	end
	if noData then return pairs(dummy) end --ensures we don't reach the next check on initialized tables
	if page < 1 or page > guildBank.numTabs then
		error(("LibGuildBankComm:IterateLinks(page) : page must be between 1 and %d inclusive"):format(guildBank.numTabs), 2)
	end
	
	local i = 0
	return function()
		i = i + 1
		if i <= 98 and guildBank[page].links[i] then
			return i, guildBank[page].links[i], guildBank[page].stacks[i]
		end
	end
end

--Returns an iterator over the saved links of the specified page
function lib:IteratePageLinks(page)
	if type(page) ~= "number" then
		error("LibGuildBankComm:IterateLinks(page) expected number", 2)
	end
	if noData then return pairs(dummy) end
	if page < 1 or page > guildBank.numTabs then
		error(("LibGuildBankComm:IterateLinks(page) : page must be between 1 and %d inclusive"):format(guildBank.numTabs), 2)
	end
	return pairs(guildBank[page].links)
end

--Returns an iterator over the saved stack counts of the specified page
function lib:IteratePageStacks(page)
	if type(page) ~= "number" then
		error("LibGuildBankComm:IterateStacks(page) expected number", 2)
	end
	if noData then return pairs(dummy) end
	if page < 1 or page > guildBank.numTabs then
		error(("LibGuildBankComm:IterateStacks(page) : page must be between 1 and %d inclusive"):format(guildBank.numTabs), 2)
	end
	return pairs(guildBank[page].stacks)
end

--Returns the requested page's scan timestamp
function lib:GetPageTimestamp(page)
	if type(page) ~= "number" then
		error("LibGuildBankComm:GetPageTimestamp(page) expected number", 2)
	end
	if noData then return 0 end
	if page < 1 or page > guildBank.numTabs then
		error(("LibGuildBankComm:GetPageTimestamp(page) : page must be between 1 and %d inclusive"):format(guildBank.numTabs), 2)
	end
	return guildBank[page] and guildBank[page].lastScan or 0
end

function lib:Start()
	lib.eventFrame:RegisterEvent("PLAYER_GUILD_UPDATE")
	lib.eventFrame:RegisterEvent("GUILD_RANKS_UPDATE")
	lib.eventFrame:RegisterEvent("GUILDBANKBAGSLOTS_CHANGED")
	lib.eventFrame:RegisterEvent("GUILDBANKFRAME_OPENED")
	lib.eventFrame:RegisterEvent("GUILDBANK_UPDATE_TABS")
	lib.eventFrame:RegisterEvent("GUILDBANK_UPDATE_MONEY")	
	lib.eventFrame:RegisterEvent("PLAYER_LOGIN")
	lib.eventFrame:RegisterEvent("GUILD_ROSTER_UPDATE")
end
lib:Start()